/**
 * OW2 Util
 * Copyright (C) 2006-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EjbJarClassMetadataConfigurator.java 6163 2012-02-08 11:02:55Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.ejbjar.impl.configurator;

import java.util.Set;

import org.ow2.util.ee.deploy.api.deployable.EJBDeployable;
import org.ow2.util.ee.metadata.common.api.struct.IJInterceptors;
import org.ow2.util.ee.metadata.common.impl.configurator.visitor.JavaxAnnotationSecurityPermitAllVisitor;
import org.ow2.util.ee.metadata.common.impl.configurator.visitor.JavaxAnnotationSecurityRolesAllowedVisitor;
import org.ow2.util.ee.metadata.common.impl.configurator.visitor.JavaxEjbTransactionAttributeVisitor;
import org.ow2.util.ee.metadata.common.impl.configurator.visitor.JavaxInterceptorExcludeDefaultInterceptorsVisitor;
import org.ow2.util.ee.metadata.common.impl.configurator.visitor.JavaxInterceptorInterceptorsVisitor;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarClassMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarDeployableMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarFieldMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarMethodMetadata;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.JavaxEjbLockVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxAnnotationSecurityDeclareRolesVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbApplicationExceptionVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbConcurrencyManagementVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbLocalBeanVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbLocalHomeVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbLocalVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbMessageDrivenVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbRemoteHomeVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbRemoteVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbSingletonVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbStartupVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbStatefulVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbStatelessVisitor;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.visitor.clazz.JavaxEjbTransactionManagementVisitor;
import org.ow2.util.ee.metadata.ws.impl.configurator.WebservicesClassMetadataConfigurator;
import org.ow2.util.scan.api.configurator.IFieldConfigurator;
import org.ow2.util.scan.api.configurator.IMethodConfigurator;
import org.ow2.util.scan.api.metadata.structures.JField;
import org.ow2.util.scan.api.metadata.structures.JMethod;


/**
 * EjbJarClassMetadataConfigurator.
 * @param <E> an implementation of EJBDeployable
 * @param <D> an implementation of IEjbJarDeployableMetadata
 * @param <C> an implementation of IEjbJarClassMetadata
 * @param <M> an implementation of IEjbJarMethodMetadata
 * @param <F> an implementation of IEjbJarFieldMetadata
 * @author Gael Lalire
 */
public abstract class EjbJarClassMetadataConfigurator<
    E extends EJBDeployable<E>,
    D extends IEjbJarDeployableMetadata<E, D, C, M, F>,
    C extends IEjbJarClassMetadata<E, D, C, M, F>,
    M extends IEjbJarMethodMetadata<E, D, C, M, F>,
    F extends IEjbJarFieldMetadata<E, D, C, M, F>>
        extends WebservicesClassMetadataConfigurator<C, M, F> {

    /**
     * annotationParsingDesactived.
     */
    private boolean annotationParsingDesactived;

    /**
     * ejbJarArchiveMetadata.
     */
    private D ejbJarArchiveMetadata;

    /**
     * EjbJarClassMetadataConfigurator.
     * @param classMetadata the classmetadata
     * @param ejbJarArchiveMetadata parent
     * @param annotationParsingDesactived annotationParsingDesactived
     */
    public EjbJarClassMetadataConfigurator(final C classMetadata, final D ejbJarArchiveMetadata,
            final boolean annotationParsingDesactived) {
        super(classMetadata, null);
        this.ejbJarArchiveMetadata = ejbJarArchiveMetadata;
        this.annotationParsingDesactived = annotationParsingDesactived;
        if (!annotationParsingDesactived) {
            registerAnnotationVisitor();
        }
    }

    /**
     * register annotation visitor.
     */
    @Override
    protected void registerAnnotationVisitor() {

        super.registerAnnotationVisitor();

        // TODO gaellalire : use interface in visitors
        C ejbJarClassMetadata = getClassMetadata();

        // add @Local
        getAnnotationVisitors().put(JavaxEjbLocalVisitor.TYPE, new JavaxEjbLocalVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @LocalBean
        getAnnotationVisitors().put(JavaxEjbLocalBeanVisitor.TYPE, new JavaxEjbLocalBeanVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @Remote
        getAnnotationVisitors().put(JavaxEjbRemoteVisitor.TYPE, new JavaxEjbRemoteVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @Stateless
        getAnnotationVisitors().put(JavaxEjbStatelessVisitor.TYPE,
                new JavaxEjbStatelessVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @Stateful
        getAnnotationVisitors().put(JavaxEjbStatefulVisitor.TYPE,
                new JavaxEjbStatefulVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @MessageDriven
        getAnnotationVisitors().put(JavaxEjbMessageDrivenVisitor.TYPE, new JavaxEjbMessageDrivenVisitor<E, D, C, M, F>(
                ejbJarClassMetadata));

        // add @LocalHome
        getAnnotationVisitors().put(JavaxEjbLocalHomeVisitor.TYPE,
                new JavaxEjbLocalHomeVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @RemoteHome
        getAnnotationVisitors().put(JavaxEjbRemoteHomeVisitor.TYPE, new JavaxEjbRemoteHomeVisitor<E, D, C, M, F>(
                ejbJarClassMetadata));

        // add @TransactionManagement
        getAnnotationVisitors().put(JavaxEjbTransactionManagementVisitor.TYPE,
                new JavaxEjbTransactionManagementVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @TransactionAttribute
        getAnnotationVisitors().put(JavaxEjbTransactionAttributeVisitor.TYPE,
                new JavaxEjbTransactionAttributeVisitor<C>(ejbJarClassMetadata));

        // add @Interceptors
        getAnnotationVisitors().put(JavaxInterceptorInterceptorsVisitor.TYPE,
                new JavaxInterceptorInterceptorsVisitor<C>(ejbJarClassMetadata));

        // add @ApplicationException
        getAnnotationVisitors().put(JavaxEjbApplicationExceptionVisitor.TYPE,
                new JavaxEjbApplicationExceptionVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @DeclareRoles
        getAnnotationVisitors().put(JavaxAnnotationSecurityDeclareRolesVisitor.TYPE,
                new JavaxAnnotationSecurityDeclareRolesVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @RolesAllowed
        getAnnotationVisitors().put(JavaxAnnotationSecurityRolesAllowedVisitor.TYPE,
                new JavaxAnnotationSecurityRolesAllowedVisitor<C>(ejbJarClassMetadata));

        // add @PermitAll
        getAnnotationVisitors().put(JavaxAnnotationSecurityPermitAllVisitor.TYPE,
                new JavaxAnnotationSecurityPermitAllVisitor<C>(ejbJarClassMetadata));

        // add @ExcludeDefaultInterceptors
        getAnnotationVisitors().put(JavaxInterceptorExcludeDefaultInterceptorsVisitor.TYPE,
                new JavaxInterceptorExcludeDefaultInterceptorsVisitor<C>(ejbJarClassMetadata));

        // add @Singleton
        getAnnotationVisitors().put(JavaxEjbSingletonVisitor.TYPE,
                new JavaxEjbSingletonVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @Startup
        getAnnotationVisitors().put(JavaxEjbStartupVisitor.TYPE,
                new JavaxEjbStartupVisitor<E, D, C, M, F>(ejbJarClassMetadata));

        // add @Lock
        getAnnotationVisitors().put(JavaxEjbLockVisitor.TYPE,
                new JavaxEjbLockVisitor<C>(ejbJarClassMetadata));

        // add @ConcurrencyManagement
        getAnnotationVisitors().put(JavaxEjbConcurrencyManagementVisitor.TYPE,
                new JavaxEjbConcurrencyManagementVisitor<E, D, C, M, F>(ejbJarClassMetadata));
    }

    /**
     * @param jField signature
     * @return configurator
     */
    @Override
    protected IFieldConfigurator createFieldMetadataConfigurator(final JField jField) {
        return createFieldMetadataConfigurator(jField, getClassMetadata(), this.annotationParsingDesactived);
    }

    /**
     * @param jMethod signature
     * @return configurator
     */
    @Override
    protected IMethodConfigurator createMethodMetadataConfigurator(final JMethod jMethod) {
        return createMethodMetadataConfigurator(jMethod, getClassMetadata(), this.annotationParsingDesactived);
    }

    /**
     * Create a field metadataconfigurator for this class configurator.
     * @param jField the given field
     * @param classMetadata the given classmetadata
     * @param annotationParsingDesactived true if annotation needs to be analyzed
     * @return a new instance of a field configurator
     */
    protected abstract IFieldConfigurator createFieldMetadataConfigurator(final JField jField, C classMetadata,
            final boolean annotationParsingDesactived);

    /**
     * Create a method metadataconfigurator for this class configurator.
     * @param jMethod the given field
     * @param classMetadata the given classmetadata
     * @param annotationParsingDesactived true if annotation needs to be analyzed
     * @return a new instance of a field configurator
     */
    protected abstract IMethodConfigurator createMethodMetadataConfigurator(final JMethod jMethod, C classMetadata,
            final boolean annotationParsingDesactived);

    /**
     * add class to archive.
     * @param referencedClasses you may fill this set with class referenced in current metadata
     */
    @Override
    public void configurationComplete(final Set<String> referencedClasses) {
       C ejbJarClassMetadata = getClassMetadata();
       ejbJarArchiveMetadata.addScannedClassMetadata(ejbJarClassMetadata);
       // add interceptors class name on class
       IJInterceptors interceptors = ejbJarClassMetadata.getAnnotationInterceptors();
       if (interceptors != null) {
           referencedClasses.addAll(interceptors.getClasses());
       }
       for (M methodMetadata : ejbJarClassMetadata.getStandardMethodMetadataCollection()) {
           // add interceptors class name on method
           interceptors = methodMetadata.getAnnotationInterceptors();
           if (interceptors != null) {
               referencedClasses.addAll(interceptors.getClasses());
           }
       }
       super.configurationComplete(referencedClasses);
    }

}
