/**
 * EasyBeans
 * Copyright (C) 2006-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EjbJarMetadataMerge.java 6193 2012-03-19 14:57:47Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.ejbjar.impl.xml.helper;

import static org.ow2.util.ee.metadata.common.api.xml.struct.ILifeCycleCallback.POST_ACTIVATE;
import static org.ow2.util.ee.metadata.common.api.xml.struct.ILifeCycleCallback.POST_CONSTRUCT;
import static org.ow2.util.ee.metadata.common.api.xml.struct.ILifeCycleCallback.PRE_DESTROY;
import static org.ow2.util.ee.metadata.common.api.xml.struct.ILifeCycleCallback.PRE_PASSIVATE;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.ejb.Remove;
import javax.ejb.TransactionAttributeType;
import javax.ejb.TransactionManagementType;
import javax.persistence.PersistenceContextType;

import org.ow2.util.asm.Type;
import org.ow2.util.ee.deploy.api.deployable.EJBDeployable;
import org.ow2.util.ee.metadata.common.api.ICommonFieldMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonMethodMetadata;
import org.ow2.util.ee.metadata.common.api.interfaces.ISharedMetadata;
import org.ow2.util.ee.metadata.common.api.struct.IEnvEntry;
import org.ow2.util.ee.metadata.common.api.struct.IJAnnotationResource;
import org.ow2.util.ee.metadata.common.api.struct.IJEjbEJB;
import org.ow2.util.ee.metadata.common.api.struct.IJInterceptors;
import org.ow2.util.ee.metadata.common.api.struct.IJavaxPersistenceContext;
import org.ow2.util.ee.metadata.common.api.struct.IJavaxPersistenceUnit;
import org.ow2.util.ee.metadata.common.api.struct.IJaxwsWebServiceRef;
import org.ow2.util.ee.metadata.common.api.struct.IJwsHandlerChain;
import org.ow2.util.ee.metadata.common.api.xml.struct.IActivationConfig;
import org.ow2.util.ee.metadata.common.api.xml.struct.IAroundInvoke;
import org.ow2.util.ee.metadata.common.api.xml.struct.ICommonEJBRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IEJBLocalRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IEJBRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IEnvironment;
import org.ow2.util.ee.metadata.common.api.xml.struct.IInjectionTarget;
import org.ow2.util.ee.metadata.common.api.xml.struct.ILifeCycleCallback;
import org.ow2.util.ee.metadata.common.api.xml.struct.IMessageDestinationRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IMethodDD;
import org.ow2.util.ee.metadata.common.api.xml.struct.IPersistenceContextRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IPersistenceUnitRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IResourceEnvRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IResourceRef;
import org.ow2.util.ee.metadata.common.api.xml.struct.IServiceRef;
import org.ow2.util.ee.metadata.common.impl.CommonFieldMetadata;
import org.ow2.util.ee.metadata.common.impl.CommonMethodMetadata;
import org.ow2.util.ee.metadata.common.impl.enc.ENCBindingException;
import org.ow2.util.ee.metadata.common.impl.struct.JAnnotationResource;
import org.ow2.util.ee.metadata.common.impl.struct.JEjbEJB;
import org.ow2.util.ee.metadata.common.impl.struct.JInterceptors;
import org.ow2.util.ee.metadata.common.impl.struct.JavaxPersistenceContext;
import org.ow2.util.ee.metadata.common.impl.struct.JavaxPersistenceUnit;
import org.ow2.util.ee.metadata.common.impl.xml.struct.ActivationConfig;
import org.ow2.util.ee.metadata.ejbjar.api.ClassType;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarClassMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarDeployableMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarFieldMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarMethodMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.TransactionAttributeLevel;
import org.ow2.util.ee.metadata.ejbjar.api.struct.IApplicationException;
import org.ow2.util.ee.metadata.ejbjar.api.struct.IJCommonBean;
import org.ow2.util.ee.metadata.ejbjar.api.struct.IJLocal;
import org.ow2.util.ee.metadata.ejbjar.api.struct.IJMessageDriven;
import org.ow2.util.ee.metadata.ejbjar.api.struct.IJRemote;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IAssemblyDescriptor;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IBean;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IContainerTransaction;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IEJB3;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IEnterpriseBeans;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IInterceptor;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IInterceptorBinding;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IInterceptors;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IMessageDriven;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IMethodPermission;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IRemoveMethod;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.ISession;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JActivationConfigProperty;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JCommonBean;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JLocal;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JMessageDriven;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JRemote;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JRemove;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.metadata.structures.JMethod;

/**
 * This class merge the information of the struct representing EJB DD into the
 * metadata.
 * @param <E> an implementation of EJBDeployable
 * @param <D> an implementation of IEjbJarDeployableMetadata
 * @param <C> an implementation of IEjbJarClassMetadata
 * @param <M> an implementation of IEjbJarMethodMetadata
 * @param <F> an implementation of IEjbJarFieldMetadata
 * @author Florent Benoit
 */
public final class EjbJarMetadataMerge<
    E extends EJBDeployable<E>,
    D extends IEjbJarDeployableMetadata<E, D, C, M, F>,
    C extends IEjbJarClassMetadata<E, D, C, M, F>,
    M extends IEjbJarMethodMetadata<E, D, C, M, F>,
    F extends IEjbJarFieldMetadata<E, D, C, M, F>> extends CommonMetadataMerge<E, D, C, M, F> {

    /**
     * Wildcard attribute.
     */
    private static final String WILDCARD = "*";

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(EjbJarMetadataMerge.class);

    /**
     * Helper class, no public constructor.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    private EjbJarMetadataMerge(final D ejbMetadata) {
        super(ejbMetadata);
    }

    /**
     * Takes struct of metadata and adds/set information on class/methods
     * metadata.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     * @param <E> an implementation of EJBDeployable
     * @param <D> an implementation of IEjbJarDeployableMetadata
     * @param <C> an implementation of IEjbJarClassMetadata
     * @param <M> an implementation of IEjbJarMethodMetadata
     * @param <F> an implementation of IEjbJarFieldMetadata
     */
    public static <E extends EJBDeployable<E>,
        D extends IEjbJarDeployableMetadata<E, D, C, M, F>,
        C extends IEjbJarClassMetadata<E, D, C, M, F>,
        M extends IEjbJarMethodMetadata<E, D, C, M, F>,
        F extends IEjbJarFieldMetadata<E, D, C, M, F>> void merge(final D ejbMetadata) {
            new EjbJarMetadataMerge<E, D, C, M, F>(ejbMetadata).resolve();
    }

    /**
     * Do all merging operations.
     */
    private void resolve() {

        IEJB3 ejb3 = getEjbJarDeployableMetadata().getEjb3();

        // First, if there are beans defined in the XML file, keep the mapping
        // EJB-NAME <--> EJB-CLASS and then add the annotated bean classes only
        // if it is the same match else ignore the annotated class
        Map<String, String> ejbNameClasses = new HashMap<String, String>();
        if (ejb3 != null) {
            // Get enterprise beans
            IEnterpriseBeans enterpriseBeans = ejb3.getEnterpriseBeans();
            if (enterpriseBeans != null) {

                // get session beans
                List<ISession> sessionList = enterpriseBeans.getSessionList();
                for (ISession session : sessionList) {
                    // get ejb-name
                    String ejbName = session.getEjbName();
                    String ejbClass = session.getEjbClass();
                    if (ejbName != null && !"".equals(ejbName)) {
                        if (ejbClass != null && !"".equals(ejbClass)) {
                            ejbNameClasses.put(ejbName, ejbClass);
                        }
                    }
                }
                // get MDBs
                List<IMessageDriven> mdbList = enterpriseBeans.getMessageDrivenList();
                for (IMessageDriven mdb : mdbList) {
                    // get ejb-name
                    String ejbName = mdb.getEjbName();
                    String ejbClass = mdb.getEjbClass();
                    if (ejbName != null && !"".equals(ejbName)) {
                        if (ejbClass != null && !"".equals(ejbClass)) {
                            ejbNameClasses.put(ejbName, ejbClass);
                        }
                    }
                }

            }
        }

        // For each class that is a bean, add a configuration for it only if
        // it's matching the XML file or there is no bean defined by XML
        for (C classMetadata : getEjbJarDeployableMetadata().getEjbJarClassMetadataCollection()) {
            if (classMetadata.isBean()) {
                String ejbName = classMetadata.getJCommonBean().getName();
                String ejbClass = classMetadata.getClassName();
                String existingEjbClassForEJBName = ejbNameClasses.get(ejbName);

                // Add it only if not defined in XML or it the matching is correct.
                if (existingEjbClassForEJBName == null || ejbClass.equals(existingEjbClassForEJBName)) {
                    getEjbJarDeployableMetadata().getBeanEjbJarClassMetadataForEjbName(ejbName, ejbClass);
                }
            }
        }

        if (ejb3 != null) {
            // Get enterprise beans
            IEnterpriseBeans enterpriseBeans = ejb3.getEnterpriseBeans();
            if (enterpriseBeans != null) {

                // get session beans
                List<ISession> sessionList = enterpriseBeans.getSessionList();
                for (ISession session : sessionList) {
                    // get ejb-name
                    String ejbName = session.getEjbName();
                    String ejbClass = session.getEjbClass();
                    C classAnnotationMetadata = getEjbJarDeployableMetadata().getBeanEjbJarClassMetadataForEjbName(ejbName,
                            ejbClass);

                    // class is found, then apply all customization.
                    applySessionBean(session, classAnnotationMetadata);
                }

                // MDB
                List<IMessageDriven> messageDrivenList = enterpriseBeans.getMessageDrivenList();
                for (IMessageDriven messageDriven : messageDrivenList) {
                    // get ejb-name
                    String ejbName = messageDriven.getEjbName();
                    String ejbClass = messageDriven.getEjbClass();
                    C classAnnotationMetadata = getEjbJarDeployableMetadata().getBeanEjbJarClassMetadataForEjbName(ejbName,
                            ejbClass);

                    // class is found, then apply all customization.
                    applyMessageDrivenBean(messageDriven, classAnnotationMetadata);
                }

            }

            // Interceptors
            IInterceptors interceptors = ejb3.getInterceptors();
            if (interceptors != null) {
                applyInterceptors(interceptors);
            }

            // Assembly descriptor
            IAssemblyDescriptor assemblyDescriptor = ejb3.getAssemblyDescriptor();
            if (assemblyDescriptor != null) {

                // merge application-exception
                mergeApplicationException(assemblyDescriptor);

                // merge interceptor-binding
                mergeInterceptorBinding(assemblyDescriptor);

                // merge container-transaction
                mergeContainerTransaction(assemblyDescriptor);

                // merge security
                mergeAssemblySecurity(assemblyDescriptor);

            }
        }
    }

    /**
     * Apply rules for the interceptors defined in the XML DD.
     * @param interceptors the interceptors struct representing
     *        &lt;interceptors&gt; element
     */
    private void applyInterceptors(final IInterceptors interceptors) {
        // Get interceptors
        List<IInterceptor> interceptorList = interceptors.getInterceptorList();
        if (interceptorList != null) {
            for (IInterceptor interceptor : interceptorList) {

                // Check that there is an interceptor class defined
                String itCptClass = interceptor.getInterceptorClass();
                if (itCptClass == null) {
                    throw new IllegalStateException("Invalid interceptor with name '" + interceptor + "': No interceptor-class");
                }

                // Class Metadata
                C itcpMetadata = getEjbJarDeployableMetadata().getScannedClassMetadata(encode(itCptClass));
                if (itcpMetadata == null) {
                    throw new IllegalStateException("No metadata found for class '" + itCptClass + "'.");
                }

                // Will use metadata of the interceptor class
                applyJndiEnvironmentRefsGroup(interceptor, itcpMetadata);

            }
        }
    }

    /**
     * Apply rules for a session bean.
     * @param sessionbean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applySessionBean(final ISession sessionbean, final C classAnnotationMetadata) {
        applyCommonBean(sessionbean, classAnnotationMetadata);

        // set session-type
        String sessionType = sessionbean.getSessionType();
        if (sessionType == null && !classAnnotationMetadata.isSession()) {
            throw new IllegalStateException("Missing session-type for bean '" + sessionbean + "' and no annotation in '"
                    + classAnnotationMetadata + "'.");
        }
        // set the type of bean
        if ("Stateless".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.STATELESS);
        } else if ("Stateful".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.STATEFUL);
        } else if ("Singleton".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.SINGLETON);
        }

        // Remote Home
        String remoteHome = sessionbean.getRemoteHome();
        if (remoteHome != null) {
            classAnnotationMetadata.setRemoteHome(remoteHome);
        }

        // Local Home
        String localHome = sessionbean.getLocalHome();
        if (localHome != null) {
            classAnnotationMetadata.setLocalHome(localHome);
        }

        // local interface
        if (sessionbean.getBusinessLocalList().size() > 0) {
            IJLocal jLocal = classAnnotationMetadata.getLocalInterfaces();
            if (jLocal == null) {
                jLocal = new JLocal();
                classAnnotationMetadata.setLocalInterfaces(jLocal);
            }
            for (String itf : sessionbean.getBusinessLocalList()) {
                String encodedItf = encode(itf);
                if (!jLocal.contains(encodedItf)) {
                    jLocal.addInterface(encodedItf);
                }
            }
        }

        // Mapped name
        if (sessionbean.getMappedName() != null) {
            IJCommonBean jCommonBean = classAnnotationMetadata.getJCommonBean();
            if (jCommonBean == null) {
                throw new IllegalStateException("No JCommonBean attribute on class '" + classAnnotationMetadata.getClassName()
                        + "' while there is a mapped name.");
            }
            jCommonBean.setMappedName(sessionbean.getMappedName());
        }

        // remote interface
        if (sessionbean.getBusinessRemoteList().size() > 0) {
            IJRemote jRemote = classAnnotationMetadata.getRemoteInterfaces();
            if (jRemote == null) {
                jRemote = new JRemote();
                classAnnotationMetadata.setRemoteInterfaces(jRemote);
            }
            for (String itf : sessionbean.getBusinessRemoteList()) {
                String encodedItf = encode(itf);
                if (!jRemote.contains(encodedItf)) {
                    jRemote.addInterface(encodedItf);
                }
            }
        }

        // Remove methods (for stateful bean)
        if (sessionbean.getRemoveMethodList().size() > 0) {
            if (!classAnnotationMetadata.isStateful()) {
                throw new IllegalStateException(
                        "Cannot apply XML DD remove methods on a bean which is not a stateful bean. Bean class is '"
                                + classAnnotationMetadata.getClassName() + "'. Remove Methods are '"
                                + sessionbean.getRemoveMethodList() + "'.");
            }

            for (IRemoveMethod removeMethod : sessionbean.getRemoveMethodList()) {
                IMethodDD method = removeMethod.getMethod();

                // method case
                List<M> methods = getMethodsForGivenMethodDD(method, classAnnotationMetadata);
                // No method found (warning)
                if (methods.size() == 0) {
                    logger.warn("No method found for the remove-method ''{0}'' on the class ''{1}''", method,
                            classAnnotationMetadata.getClassName());
                } else if (methods.size() > 1) {
                    logger.warn("Found more than 1 method for the remove-method ''{0}'' on the class ''{1}''."
                            + "All these methods will be set as remove methods.", method, classAnnotationMetadata
                            .getClassName());
                }

                // Set as a remove method
                for (M methodAnnotationMetadata : methods) {

                    // Get existing value
                    Remove previousRemove = methodAnnotationMetadata.getJRemove();

                    // Value of the XML (for retain)
                    Boolean retainXML = removeMethod.getRetainIfException();

                    // If value has been defined in the XML, use this value,
                    // else use the value of the annotation, else the default
                    // value (false)
                    boolean retain = false;
                    if (retainXML != null) {
                        retain = retainXML.booleanValue();
                    } else if (previousRemove != null) {
                        // got value of the annotation
                        retain = previousRemove.retainIfException();
                    }

                    // Set the new value
                    JRemove jRemove = new JRemove(retain);
                    methodAnnotationMetadata.setRemove(jRemove);
                }
            }
        }

    }


    /**
     * Apply rules for a Message Driven Bean.
     * @param messageDrivenBean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyMessageDrivenBean(final IMessageDriven messageDrivenBean,
            final C classAnnotationMetadata) {
        applyCommonBean(messageDrivenBean, classAnnotationMetadata);

        // Message Destination type is set ? add an activation config property
        if (messageDrivenBean.getMessageDestinationType() != null) {
            ActivationConfig activationConfig = new ActivationConfig();
            messageDrivenBean.addActivationConfig(activationConfig);
            activationConfig.setName("destinationType");
            activationConfig.setValue(messageDrivenBean.getMessageDestinationType());
        }

        // Set MDB type
        classAnnotationMetadata.setClassType(ClassType.MDB);

        // Init MessageDriven object
        IJMessageDriven jMessageDriven = classAnnotationMetadata.getJMessageDriven();
        if (jMessageDriven == null) {
            jMessageDriven = new JMessageDriven();
            classAnnotationMetadata.setJMessageDriven(jMessageDriven);
        }

        // Messaging type
        String messagingType = messageDrivenBean.getMessagingType();
        if (messagingType != null) {
            String existingInterface = jMessageDriven.getMessageListenerInterface();
            if (existingInterface == null) {
                jMessageDriven.setMessageListenerInterface(messagingType);
            }
        }

        // Activation config
        List<IActivationConfig> activationConfigList = messageDrivenBean.getActivationConfigList();
        for (IActivationConfig activationConfig : activationConfigList) {
            JActivationConfigProperty actProperty = new JActivationConfigProperty(activationConfig.getName(), activationConfig
                    .getValue());
            jMessageDriven.addActivationConfigProperty(actProperty);
        }

        // Message destination link ?
        jMessageDriven.setMessageDestinationLink(messageDrivenBean.getMessageDestinationLink());

    }

    /**
     * Apply rules for a common bean.
     * @param bean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyCommonBean(final IBean bean, final C classAnnotationMetadata) {
        // no Common bean, create one.
        if (classAnnotationMetadata.getJCommonBean() == null) {
            classAnnotationMetadata.setJCommonBean(new JCommonBean());
        }
        // Sets the ejb-name
        IJCommonBean jCommonBean = classAnnotationMetadata.getJCommonBean();
        jCommonBean.setName(bean.getEjbName());

        // Mapped name
        if (bean.getMappedName() != null) {
            jCommonBean.setMappedName(bean.getMappedName());
        }

        // Apply run-as role
        String runAsRole = bean.getRunAsRole();
        if (runAsRole != null) {
            classAnnotationMetadata.setRunAs(runAsRole);
        }

        // Add security-role-ref list to the metadata
        classAnnotationMetadata.setSecurityRoleRefList(bean.getSecurityRoleRefList());

        // Transaction type
        String transactionType = bean.getTransactionType();
        if (transactionType != null) {
            if ("Bean".equals(transactionType)) {
                classAnnotationMetadata.setTransactionManagementType(TransactionManagementType.BEAN);
            } else if ("Container".equals(transactionType)) {
                classAnnotationMetadata.setTransactionManagementType(TransactionManagementType.CONTAINER);
            } else {
                throw new IllegalStateException("Invalid transaction type '" + transactionType + "' found");
            }
        }

        applyJndiEnvironmentRefsGroup(bean, classAnnotationMetadata);

    }

    /**
     * Apply rules for a an environment.
     * @param envGroup the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    protected void applyJndiEnvironmentRefsGroup(final IEnvironment envGroup,
            final C classAnnotationMetadata) {

        // post-construct / pre-destroy
        applyLifeCycle(envGroup.getPostConstructCallbackList(), classAnnotationMetadata, POST_CONSTRUCT);
        applyLifeCycle(envGroup.getPreDestroyCallbackList(), classAnnotationMetadata, PRE_DESTROY);

        // pre-passivate / post-activate / aroundInvoke
        applyLifeCycle(envGroup.getPrePassivateCallbackList(), classAnnotationMetadata, PRE_PASSIVATE);
        applyLifeCycle(envGroup.getPostActivateCallbackList(), classAnnotationMetadata, POST_ACTIVATE);
        applyAroundInvoke(envGroup.getAroundInvokeList(), classAnnotationMetadata, IAroundInvoke.NAME);

        // ejb(local) ref
        applyEjbRef(envGroup.getEJBRefList(), classAnnotationMetadata);
        applyEjbLocalRef(envGroup.getEJBLocalRefList(), classAnnotationMetadata);

        // resource(env) ref
        applyResourceRef(envGroup.getResourceRefList(), classAnnotationMetadata);
        applyResourceEnvRef(envGroup.getResourceEnvRefList(), classAnnotationMetadata);

        // env entry
        applyEnvEntry(envGroup.getEnvEntryList(), classAnnotationMetadata);

        // persistence unit ref
        applyPersistenceUnitRef(envGroup.getPersistenceUnitRefList(), classAnnotationMetadata);

        // persistence context ref
        applyPersistenceContextRef(envGroup.getPersistenceContextRefList(), classAnnotationMetadata);

        // message-destination-ref
        applyMessageDestinationRef(envGroup.getMessageDestinationRefList(), classAnnotationMetadata);

        // service-ref
        applyServiceRef(envGroup.getServiceRefList(), classAnnotationMetadata);

        // handler-chains
        applyHandlerChain(classAnnotationMetadata);

    }


    /**
     * Apply rules for a common bean (env-entry).
     * @param envEntryList the list of env-entry.
     * @param classAnnotationMetadata (Annotation metadata) of the bean
     */
    protected void applyEnvEntry(final List<org.ow2.util.ee.metadata.common.api.xml.struct.IEnvEntry> envEntryList,
            final C classAnnotationMetadata) {
        if (envEntryList == null) {
            return;
        }

        List<IEnvEntry> structEnvEntryList = new ArrayList<IEnvEntry>();

        // Store the env-entry list in the metadata class
        for (org.ow2.util.ee.metadata.common.api.xml.struct.IEnvEntry envEntry : envEntryList) {
            org.ow2.util.ee.metadata.common.impl.struct.EnvEntry builtEntry = new org.ow2.util.ee.metadata.common.impl.struct.EnvEntry(envEntry.getEnvEntryName(), envEntry
                    .getEnvEntryType(), envEntry.getEnvEntryValue(), envEntry.getLookupName());
            structEnvEntryList.add(builtEntry);

            // for each injection-target field, add a resource annotation
            List<IInjectionTarget> injectionTargetList = envEntry.getInjectionTargetList();
            if (injectionTargetList != null) {
                for (IInjectionTarget injectionTarget : injectionTargetList) {

                    // Search injection object
                    ISharedMetadata sharedMetadata = getInjectionTarget(classAnnotationMetadata, injectionTarget);
                    if (builtEntry.getType() == null) {
                        builtEntry.setType(getEnvEntryType(sharedMetadata));
                    }
                    
                    // build a new annotation only if there is a value specified or if lookupName is specified
                    if (envEntry.getEnvEntryValue() == null && envEntry.getLookupName() == null) {
                        continue;
                    }

                    JAnnotationResource jResource = new JAnnotationResource();

                    // set name
                    jResource.setName(envEntry.getEnvEntryName());

                    // set type
                    jResource.setType(envEntry.getEnvEntryType());
                    
                    // Add field injection
                    jResource.setLookup(envEntry.getLookupName());

                    // Add the JAnnotationResource
                    sharedMetadata.setJAnnotationResource(jResource);
                    
                }
            }



            //applyJResource(jResource, classAnnotationMetadata, resourceEnvRef.getInjectionTargetList());
        }

        // Set the env entry collection
        classAnnotationMetadata.setEnvEntryCollection(structEnvEntryList);

    }

    /**
     * Apply rules for a common bean (service-ref).
     * @param serviceRefList the list of service-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    protected void applyServiceRef(final List<IServiceRef> serviceRefList,
                                   final C classAnnotationMetadata) {
        if (serviceRefList == null) {
            return;
        }

        // For each service-ref, build a &#64;{@link
        // javax.xml.ws.WebServiceRef}
        for (IServiceRef serviceRef : serviceRefList) {

            List<IInjectionTarget> injectionTargetList = serviceRef.getInjectionTargetList();

            // Honor the injection-target if specified
            if (injectionTargetList != null && injectionTargetList.size() > 0) {
                for (IInjectionTarget target : injectionTargetList) {
                    // Find the attribute / target
                    ISharedMetadata sharedMetadata = getInjectionTarget(classAnnotationMetadata, target);

                    // If there is already a WebServiceRef annotation, overload
                    // it
                    IJaxwsWebServiceRef wsr = sharedMetadata.getJaxwsWebServiceRef();

                    if (wsr != null) {
                        // merge
                        mergeWebServiceRef(serviceRef, wsr);
                    } else {
                        // place a new annotation on the target
                        sharedMetadata.setJaxwsWebServiceRef(createJaxwsWebServiceRefFromXML(serviceRef));
                    }
                }
            } else {
                // We need to lookup the target annotation in the class
                // and overload it.
                String name = serviceRef.getName();
                List<IJaxwsWebServiceRef> metadatas = findJaxwsWebServiceRefMetadata(classAnnotationMetadata, name);

                if (metadatas.isEmpty()) {
                    // no annotation to override:
                    // -> equivalent to class-level annotation

                    // If there is an existing single class-level annotation,
                    // place it in the list
                    IJaxwsWebServiceRef existing = classAnnotationMetadata.getJaxwsWebServiceRef();
                    List<IJaxwsWebServiceRef> webServiceRefs = classAnnotationMetadata.getJaxwsWebServiceRefs();

                    // init the list if needed
                    if (webServiceRefs == null) {
                        webServiceRefs = new ArrayList<IJaxwsWebServiceRef>();
                        classAnnotationMetadata.setJaxwsWebServiceRefs(webServiceRefs);
                    }

                    if (existing != null) {
                        // place the annotation in the @WebServiceRefs list
                        webServiceRefs.add(existing);
                        // remove the single @WebServiceRef
                        classAnnotationMetadata.setJaxwsWebServiceRef(null);
                    }

                    // add our new annotation in the list
                    webServiceRefs.add(createJaxwsWebServiceRefFromXML(serviceRef));

                } else if (metadatas.size() > 1) {
                    // Not good, too many choices
                    // a smart tool may try to find a compatible existing
                    // metadata and perform the merge
                    // and a dumb tool will produce an Exception :)
                    throw new IllegalStateException("Too many matching WebServiceRef metadata for '" + name
                            + "' service-ref of class '" + classAnnotationMetadata.getClassName() + "'");
                } else {
                    // only 1 matching candidate
                    // merge
                    mergeWebServiceRef(serviceRef, metadatas.get(0));
                }
            }
        }
    }

    /**
     * Needs to search the matching WebServiceRef.
     * @param classAnnotationMetadata the metadata
     * @param name the xml name of the service ref
     * @return a list of matching web service ref
     */
    private List<IJaxwsWebServiceRef> findJaxwsWebServiceRefMetadata(final C classAnnotationMetadata,
            final String name) {

        List<IJaxwsWebServiceRef> elements = new ArrayList<IJaxwsWebServiceRef>();

        // Find annotation on Class
        // Try @WebServiceRef
        IJaxwsWebServiceRef classAnnotation = classAnnotationMetadata.getJaxwsWebServiceRef();
        if ((classAnnotation != null) && (name.equals(classAnnotation.getName()))) {
            elements.add(classAnnotation);
        }
        // Try @WebServiceRefs
        List<IJaxwsWebServiceRef> webServiceRefs = classAnnotationMetadata.getJaxwsWebServiceRefs();
        if (webServiceRefs != null) {
            for (IJaxwsWebServiceRef ref : webServiceRefs) {
                if (name.equals(ref.getName())) {
                    elements.add(ref);
                }
            }
        }

        // search on fields
        Collection<? extends IEjbJarFieldMetadata<?, ?, ?, ?, ?>> fieldMetadatas = classAnnotationMetadata
                .getStandardFieldMetadataCollection();
        if (fieldMetadatas != null) {
            for (IEjbJarFieldMetadata<?, ?, ?, ?, ?> fieldMetadata : fieldMetadatas) {
                addJaxwsWebServiceRefIfFound(fieldMetadata, name, elements);
            }
        }
        Collection<? extends IEjbJarMethodMetadata<?, ?, ?, ?, ?>> methodMetadatas = classAnnotationMetadata
                .getStandardMethodMetadataCollection();
        if (methodMetadatas != null) {
            for (IEjbJarMethodMetadata<?, ?, ?, ?, ?> methodMetadata : methodMetadatas) {
                addJaxwsWebServiceRefIfFound(methodMetadata, name, elements);
            }
        }



        return elements;
    }

    /**
     * Apply rules for a common bean (message-destination-ref).
     * @param messageDestinationRefList the list of message-destination-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyMessageDestinationRef(final List<IMessageDestinationRef> messageDestinationRefList,
            final C classAnnotationMetadata) {
        if (messageDestinationRefList == null) {
            return;
        }

        // For each message-destination-ref, build a &#64;{@link
        // javax.annotation.Resource}
        for (IMessageDestinationRef messageDestinationRef : messageDestinationRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(messageDestinationRef.getName());

            // set type
            jResource.setType(messageDestinationRef.getType());

            // Mapped Name
            jResource.setMappedName(messageDestinationRef.getMappedName());

            // sets the lookup-name
            if (messageDestinationRef.getLookupName() != null) {
                jResource.setLookup(messageDestinationRef.getLookupName());
            }         
            
            // Link ?
            jResource.setMessageDestinationLink(messageDestinationRef.getLink());

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, messageDestinationRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (resource-ref).
     * @param resourceRefList the list of resource-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyResourceRef(final List<IResourceRef> resourceRefList, final C
            classAnnotationMetadata) {
        if (resourceRefList == null) {
            return;
        }

        // For each resource-ref, build a &#64;{@link javax.annotation.Resource}
        for (IResourceRef resourceRef : resourceRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(resourceRef.getResRefName());

            // set type
            jResource.setType(resourceRef.getResRefType());

            // Mapped Name
            jResource.setMappedName(resourceRef.getMappedName());
            
            // sets the lookup-name
            if (resourceRef.getLookupName() != null) {
                jResource.setLookup(resourceRef.getLookupName());
            }       

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, resourceRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (resource-env-ref).
     * @param resourceEnvRefList the list of resource-env-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyResourceEnvRef(final List<IResourceEnvRef> resourceEnvRefList,
            final C classAnnotationMetadata) {
        if (resourceEnvRefList == null) {
            return;
        }

        // For each resource-ref, build a &#64;{@link javax.annotation.Resource}
        for (IResourceEnvRef resourceEnvRef : resourceEnvRefList) {
            JAnnotationResource jResource = new JAnnotationResource();
            // set name
            jResource.setName(resourceEnvRef.getResourceEnvRefName());

            // set type
            jResource.setType(resourceEnvRef.getResourceEnvRefType());

            // Mapped Name
            jResource.setMappedName(resourceEnvRef.getMappedName());
            
            // sets the lookup-name
            if (resourceEnvRef.getLookupName() != null) {
                jResource.setLookup(resourceEnvRef.getLookupName());
            }

            // Apply the resource on the class
            applyJResource(jResource, classAnnotationMetadata, resourceEnvRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (persistence-unit-ref).
     * @param persistenceUnitRefList the list of persistence-unit-ref
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null)
     */
    private void applyPersistenceUnitRef(final List<IPersistenceUnitRef> persistenceUnitRefList,
            final C classAnnotationMetadata) {
        if (persistenceUnitRefList == null) {
            return;
        }

        // For each persistence-unit-ref, build a &#64;{@link javax.persistence.PersistenceUni}
        for (IPersistenceUnitRef persistenceUnitRef : persistenceUnitRefList) {
            JavaxPersistenceUnit jPersistenceUnit =  new JavaxPersistenceUnit();
            // set ref-name
            jPersistenceUnit.setName(persistenceUnitRef.getPersistenceUnitRefName());

            // set unit-name
            jPersistenceUnit.setUnitName(persistenceUnitRef.getPersistenceUnitName());

            applyJPersistenceUnit(jPersistenceUnit, classAnnotationMetadata, persistenceUnitRef.getInjectionTargetList());
        }
    }

    private void applyPersistenceContextRef(final List<IPersistenceContextRef> persistenceContextRefList,
            final C classAnnotationMetadata) {
        if (persistenceContextRefList == null) {
            return;
        }

        // For each persistence-unit-ref, build a &#64;{@link javax.persistence.PersistenceUni}
        for (IPersistenceContextRef persistenceContextRef : persistenceContextRefList) {
            JavaxPersistenceContext jPersistenceContext =  new JavaxPersistenceContext();
            // set ref-name
            jPersistenceContext.setName(persistenceContextRef.getPersistenceContextRefName());

            // set unit-name
            jPersistenceContext.setUnitName(persistenceContextRef.getPersistenceUnitName());

            // set context type
            String type = persistenceContextRef.getPersistenceUnitName();
            if (PersistenceContextType.TRANSACTION.toString().equals(type)) {
                jPersistenceContext.setType(PersistenceContextType.TRANSACTION);
            } else if (PersistenceContextType.EXTENDED.toString().equals(type)) {
                jPersistenceContext.setType(PersistenceContextType.EXTENDED);
            }

            applyJPersistenceContext(jPersistenceContext, classAnnotationMetadata, persistenceContextRef.getInjectionTargetList());
        }
    }
    /**
     * Adds the given &#64;{@link javax.persistence.PersistenceUnit} object to the
     * class and apply it also on injection target elements.
     * @param jPersistenceUnit an &#64;{@link javax.persistence.PersistenceUnit} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    protected void applyJPersistenceUnit(final JavaxPersistenceUnit jPersistenceUnit,
            final C classAnnotationMetadata, final List<IInjectionTarget> injectionTargetList) {

        // Add field on the class
        IJavaxPersistenceUnit existingJPersistentUnit = classAnnotationMetadata.getJavaxPersistenceUnit();

        // Needs to use a list instead.
        List<IJavaxPersistenceUnit> jPersistentUnits = classAnnotationMetadata.getJavaxPersistencePersistenceUnits();
        if (jPersistentUnits == null) {
            jPersistentUnits = new ArrayList<IJavaxPersistenceUnit>();
            classAnnotationMetadata.setJavaxPersistencePersistenceUnits(jPersistentUnits);
        }

        // Put existing existing element in the list
        if (existingJPersistentUnit != null) {
            jPersistentUnits.add(existingJPersistentUnit);
            classAnnotationMetadata.setJavaxPersistenceUnit(null);
        }

        // Add the new element
        jPersistentUnits.add(jPersistenceUnit);


        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (IInjectionTarget injectionTarget : injectionTargetList) {

                // Add the JavaxPersistenceUnit on it
                getInjectionTarget(classAnnotationMetadata, injectionTarget).setJavaxPersistenceUnit(jPersistenceUnit);
            }
        }
    }
    protected void applyJPersistenceContext(final JavaxPersistenceContext jPersistenceContext,
            final C classAnnotationMetadata, final List<IInjectionTarget> injectionTargetList) {

        // Add field on the class
        IJavaxPersistenceContext existingJPersistentContext = classAnnotationMetadata.getJavaxPersistenceContext();

        // Needs to use a list instead.
        List<IJavaxPersistenceContext> jPersistentContexts = classAnnotationMetadata.getJavaxPersistencePersistenceContexts();
        if (jPersistentContexts == null) {
            jPersistentContexts = new ArrayList<IJavaxPersistenceContext>();
            classAnnotationMetadata.setJavaxPersistencePersistenceContexts(jPersistentContexts);
        }

        // Put existing existing element in the list
        if (existingJPersistentContext != null) {
            jPersistentContexts.add(existingJPersistentContext);
            classAnnotationMetadata.setJavaxPersistenceContext(null);
        }

        // Add the new element
        jPersistentContexts.add(jPersistenceContext);


        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (IInjectionTarget injectionTarget : injectionTargetList) {

                // Add the JavaxPersistenceContext on it
                getInjectionTarget(classAnnotationMetadata, injectionTarget).setJavaxPersistenceContext(jPersistenceContext);
            }
        }
    }
    /**
     * Apply rules for a common bean (ejb-ref).
     * @param ejbRefList the list of ejb-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyEjbRef(final List<IEJBRef> ejbRefList, final C classAnnotationMetadata) {
        if (ejbRefList == null) {
            return;
        }

        // For each ejb-ref, build a &#64;{@link javax.ejb.EJB}
        for (IEJBRef ejbRef : ejbRefList) {
            JEjbEJB jEJB = buildEJB(ejbRef);
            // set interface
            if (ejbRef.getHome() != null) {
                jEJB.setBeanInterface(ejbRef.getHome());
            } else if (ejbRef.getRemote() != null) {
                jEJB.setBeanInterface(ejbRef.getRemote());
            }
            applyJEJB(jEJB, classAnnotationMetadata, ejbRef.getInjectionTargetList());
        }
    }

    /**
     * Apply rules for a common bean (ejb-local-ref).
     * @param ejbLocalRefList the list of ejb-local-ref.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     */
    private void applyEjbLocalRef(final List<IEJBLocalRef> ejbLocalRefList, final C classAnnotationMetadata) {
        if (ejbLocalRefList == null) {
            return;
        }

        // For each ejb-local-ref, build a &#64;{@link javax.ejb.EJB}
        for (IEJBLocalRef ejbLocalRef : ejbLocalRefList) {
            JEjbEJB jEJB = buildEJB(ejbLocalRef);
            // set interface
            if (ejbLocalRef.getLocalHome() != null) {
                jEJB.setBeanInterface(ejbLocalRef.getLocalHome());
            } else if (ejbLocalRef.getLocal() != null) {
                jEJB.setBeanInterface(ejbLocalRef.getLocal());
            }
            applyJEJB(jEJB, classAnnotationMetadata, ejbLocalRef.getInjectionTargetList());
        }
    }

    /**
     * Adds the given &#64;{@link javax.annotation.Resource} object to the
     * class and apply it also on injection target elements.
     * @param jResource an &#64;{@link javax.annotation.Resource} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    protected void applyJResource(final JAnnotationResource jResource,
            final C classAnnotationMetadata, final List<IInjectionTarget> injectionTargetList) {

        // Add field on the class
        IJAnnotationResource existingJResource = classAnnotationMetadata.getJAnnotationResource();

        // Needs to use a list instead.
        List<IJAnnotationResource> jResources = classAnnotationMetadata.getJAnnotationResources();
        if (jResources == null) {
            jResources = new ArrayList<IJAnnotationResource>();
            classAnnotationMetadata.setJAnnotationResources(jResources);
        }

        // put existing element in the list
        if (existingJResource != null) {
            jResources.add(existingJResource);
            // remove single element
            classAnnotationMetadata.setJAnnotationResource(null);
        }

        // Add our new element (or replace/update if there is a previous entry)
        IJAnnotationResource resourceToUse = null;
        for (IJAnnotationResource currentResource : jResources) {
            if (jResource.getName().equals(currentResource.getName())) {
                resourceToUse = currentResource;
            }
        }
        if (resourceToUse != null) {
            // found a previous entry, update it
            if (jResource.getMappedName() != null && !"".equals(jResource.getMappedName())) {
                resourceToUse.setMappedName(jResource.getMappedName());
            }
        } else {
            // add it
            jResources.add(jResource);
            resourceToUse = jResource;
        }

        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (IInjectionTarget injectionTarget : injectionTargetList) {

                // Add the JAnnotationResource on it
                getInjectionTarget(classAnnotationMetadata, injectionTarget).setJAnnotationResource(resourceToUse);
            }
        }

    }

    /**
     * Gets the metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a common annotation metadata if element is found
     */
    protected ISharedMetadata getInjectionTarget(final C classAnnotationMetadata,
            final IInjectionTarget injectionTarget) {
        String injectionTargetName = injectionTarget.getTargetName();

        // First, try with the method
        M method = getMethod(classAnnotationMetadata, injectionTarget);

        // found ? return it
        if (method != null) {
            return method;
        }

        // else, try with the next item which should be a field
        // Try with the field
        F field = getField(classAnnotationMetadata, injectionTarget);
        if (field == null) {
            throw new IllegalArgumentException("No method or field for injection target name '" + injectionTargetName
                    + "' found in the class '" + classAnnotationMetadata.getClassName() + "'.");
        }

        return field;

    }

    /**
     * Gets the method metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a method metadata if element is found
     */
    protected M getMethod(final C classAnnotationMetadata,
            final IInjectionTarget injectionTarget) {
        // get data from the injection target element
        String clName = injectionTarget.getClassname();
        if (clName == null) {
            clName = classAnnotationMetadata.getClassName();
        } else {
            clName = clName.replace(".", "/");
        }

        // Get method name which is the injection target name
        String methodName = injectionTarget.getTargetName();

        // Now, search the element in the given class
        C foundClassMetadata = classAnnotationMetadata.getLinkedClassMetadata(clName);
        if (foundClassMetadata == null) {
            throw new IllegalArgumentException("Cannot find the class metadata named '" + clName + "'.");
        }

        // Got the class metadata, now search the field
        List<? extends M> methods = foundClassMetadata.searchStandardMethodMetadata(methodName);

        if (methods.size() == 0) {
            // no method
            return null;
        } else if (methods.size() > 1) {
            // Should be only one
            throw new IllegalArgumentException("Too many methods with name '" + methodName + "' found in the class '" + clName
                    + "'.");
        }

        return methods.get(0);
    }

    /**
     * Gets the field metadata for a given injection target.
     * @param classAnnotationMetadata the class used to search the target
     *        element
     * @param injectionTarget the given target info element
     * @return a field if element is found
     */
    protected F getField(final C classAnnotationMetadata,
            final IInjectionTarget injectionTarget) {
        // get data from the injection target element
        String clName = injectionTarget.getClassname();
        if (clName == null) {
            clName = classAnnotationMetadata.getClassName();
        } else {
            clName = clName.replace(".", "/");
        }

        String fieldName = injectionTarget.getTargetName();

        // Now, search the element in the given class
        C foundClassMetadata = classAnnotationMetadata.getLinkedClassMetadata(clName);

        if (foundClassMetadata == null) {
            throw new IllegalArgumentException("Cannot find the class metadata named '" + clName + "'.");
        }

        // Got the class metadata, now search the field
        List<? extends F> fields = foundClassMetadata.searchStandardFieldMetadata(fieldName);

        if (fields.size() == 0) {
            // no field
            return null;
        } else if (fields.size() > 1) {
            // Should be only one
            throw new IllegalArgumentException("Too many fields with name '" + fieldName + "' found in the class '" + clName
                    + "'.");
        }

        return fields.get(0);
    }
    
    /**
     * Gets the type for a given metadata.
     * @param commonMetadata the given metadata
     * @return the type.
     */
    private String getEnvEntryType(final ISharedMetadata commonMetadata) {
        if (commonMetadata == null) {
            throw new IllegalArgumentException("Null is not a valid metadata");
        }


        if (commonMetadata instanceof CommonFieldMetadata<?,?,?>) {
            CommonFieldMetadata<?, ?, ?> fieldAnnotationMetadata = (CommonFieldMetadata<?, ?, ?>) commonMetadata;
            Type typeInterface = Type.getType(fieldAnnotationMetadata.getJField().getDescriptor());
            return envEntryTypeConvert(typeInterface.getClassName());
        } else if (commonMetadata instanceof CommonMethodMetadata<?,?,?>) {
            CommonMethodMetadata<?, ?, ?> methodAnnotationMetadata = (CommonMethodMetadata<?, ?, ?>) commonMetadata;

            JMethod jMethod = methodAnnotationMetadata.getJMethod();
            Type[] args = Type.getArgumentTypes(jMethod.getDescriptor());
            if (args.length != 1) {
                throw new IllegalArgumentException("Method args '" + Arrays.asList(args) + "' for method '" + jMethod
                        + "' are invalid. Length should be of 1.");
            }
            Type typeInterface = args[0];
            return envEntryTypeConvert(typeInterface.getClassName());
        }
        
        throw new IllegalArgumentException("Unknown metadata type '" + commonMetadata + "'");
        
    }
    
    /**
     * Convert the given classname type into an env-entry type
     * @param className the given classname
     * @return the given env-entry-type
     */
    protected String envEntryTypeConvert(String className) {
        if (Boolean.TYPE.toString().equals(className)) {
            return Boolean.class.getName();
        } else if (Byte.TYPE.toString().equals(className)) {
            return Byte.class.getName();
        } else if (Character.TYPE.toString().equals(className)) {
            return Character.class.getName();
        } else if (Short.TYPE.toString().equals(className)) {
            return Short.class.getName();
        } else if (Integer.TYPE.toString().equals(className)) {
            return Integer.class.getName();
        } else if (Long.TYPE.toString().equals(className)) {
            return Long.class.getName();
        } else if (Float.TYPE.toString().equals(className)) {
            return Float.class.getName();
        } else if (Double.TYPE.toString().equals(className)) {
            return Double.class.getName();
        } else if (String.class.getName().equals(className)) {
            return String.class.getName();
        }  else if (Class.class.getName().equals(className)) {
            return Class.class.getName();
        }
        // Else default to the enum class 
        return className;
    }

    /**
     * Adds the given &#64;{@link javax.ejb.EJB} object to the class and apply
     * it also on injection target elements.
     * @param jEJB an &#64;{@link javax.ejb.EJB} impl
     * @param classAnnotationMetadata the bean's class
     * @param injectionTargetList the list of injection fields to set
     */
    protected void applyJEJB(final JEjbEJB jEJB, final C classAnnotationMetadata,
            final List<IInjectionTarget> injectionTargetList) {

        // Add field on the class
        IJEjbEJB existingJEjbEJB = classAnnotationMetadata.getJEjbEJB();

        // Needs to use a list instead.
        List<IJEjbEJB> jEjbEJBs = classAnnotationMetadata.getJEjbEJBs();
        if (jEjbEJBs == null) {
            jEjbEJBs = new ArrayList<IJEjbEJB>();
            classAnnotationMetadata.setJEjbEJBs(jEjbEJBs);
        }

        // put existing element in the list
        if (existingJEjbEJB != null) {
            jEjbEJBs.add(existingJEjbEJB);
            // remove single element
            classAnnotationMetadata.setJEjbEJB(null);
        }

        // Add our new element
        jEjbEJBs.add(jEJB);

        // Now, add elements on the injection fields
        if (injectionTargetList != null) {
            for (IInjectionTarget injectionTarget : injectionTargetList) {

                // Add the JEjbEJB on it
                getInjectionTarget(classAnnotationMetadata, injectionTarget).setJEjbEJB(jEJB);
            }
        }

    }

    /**
     * Build an &#64;{@link javax.ejb.EJB} impl for the given
     * ejb-ref/ejb-local-ref.
     * @param ref an instance of a ejb-ref/ejb-local-ref.
     * @return a implementation of &#64;{@link javax.ejb.EJB} interface.
     */
    private static JEjbEJB buildEJB(final ICommonEJBRef ref) {
        JEjbEJB jEJB = new JEjbEJB();
        // Sets the name
        if (ref.getEjbRefName() != null) {
            jEJB.setName(ref.getEjbRefName());
        }

        // sets the mapped-name
        if (ref.getMappedName() != null) {
            jEJB.setMappedName(ref.getMappedName());
        }

        // sets the lookup-name
        if (ref.getLookupName() != null) {
            jEJB.setLookup(ref.getLookupName());
        }        
        
        // sets the bean-name
        if (ref.getEjbLink() != null) {
            jEJB.setBeanName(ref.getEjbLink());
        }

        return jEJB;
    }

    /**
     * Apply rules for a common bean (method callbacks).
     * @param aroundInvokeList the list of method/aroundInvoke callbacks.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     * @param type the type of lifecycle.
     */
    private void applyAroundInvoke(final List<IAroundInvoke> aroundInvokeList,
            final C classAnnotationMetadata, final String type) {
        for (IAroundInvoke aroundInvoke : aroundInvokeList) {
            String interceptorClassName = null;

            // if no class, takes the bean's class
            if (aroundInvoke.getClassName() == null) {
                interceptorClassName = classAnnotationMetadata.getClassName();
            } else {
                interceptorClassName = encode(aroundInvoke.getClassName());
            }

            // Now, search the Method on the class or its super-class if not
            // found
            String searchMethodName = aroundInvoke.getMethodName();
            if (searchMethodName == null) {
                throw new IllegalStateException("Invalid null method name for the callback '" + interceptorClassName + "'.");
            }

            // Apply interceptor
            applyInterceptor(type, interceptorClassName, searchMethodName, classAnnotationMetadata);

        }
    }

    /**
     * Apply rules for a common bean (lifecycle callbacks).
     * @param lifecycleList the list of lifecycle callbacks.
     * @param classAnnotationMetadata (Annotation metadata) of the bean using
     *        this interceptor (may be null).
     * @param type the type of lifecycle.
     */
    private void applyLifeCycle(final List<ILifeCycleCallback> lifecycleList,
            final C classAnnotationMetadata, final String type) {
        for (ILifeCycleCallback lifecycleCallback : lifecycleList) {

            String interceptorClassName = null;

            // if no class, takes the bean's class
            if (lifecycleCallback.getLifecycleCallbackClass() == null) {
                interceptorClassName = classAnnotationMetadata.getClassName();
            } else {
                interceptorClassName = encode(lifecycleCallback.getLifecycleCallbackClass());
            }

            // Now, search the Method on the class or its super-class if not
            // found
            String searchMethodName = lifecycleCallback.getLifecycleCallbackMethod();
            if (searchMethodName == null) {
                throw new IllegalStateException("Invalid null method name for the callback '" + interceptorClassName + "'.");
            }

            // Apply interceptor
            applyInterceptor(type, interceptorClassName, searchMethodName, classAnnotationMetadata);

        }
    }

    /**
     * Adds a link on the given class to the interceptor class.
     * @param classAnnotationMetadata the class that will use/import the
     *        interceptor
     * @param interceptorClassName the interceptor class.
     */
    private void addInterceptorForClass(final C classAnnotationMetadata,
                final String interceptorClassName) {
        // Get interceptors for the given class (the bean class)
        IJInterceptors interceptors = classAnnotationMetadata.getAnnotationInterceptors();
        if (interceptors == null) {
            // build new list if empty
            interceptors = new JInterceptors();
            classAnnotationMetadata.setAnnotationsInterceptors(interceptors);
        }

        // class already present ?
        if (!interceptors.contains(interceptorClassName)) {
            interceptors.addClass(interceptorClassName);
        }
    }

    /**
     * Finds a method in the interceptor class and set the correct type value.
     * @param type type of interceptor
     * @param interceptorClassName the name of the interceptor class
     * @param methodName the name of the method in the interceptor class
     * @param classAnnotationMetadata the class that is using the interceptor
     *        (may be null)
     */
    private void applyInterceptor(final String type, final String interceptorClassName, final String methodName,
            final C classAnnotationMetadata) {

        // Name of all super classes of the bean
        List<String> superClassesOfBean = new ArrayList<String>();
        C beanClassMetadata = classAnnotationMetadata;
        // Start with the bean class name
        String superClassBeanName = beanClassMetadata.getClassName();
        while (!superClassBeanName.equals(JAVA_LANG_OBJECT)) {
            superClassesOfBean.add(superClassBeanName);
            beanClassMetadata = beanClassMetadata.getLinkedClassMetadata(superClassBeanName);
            superClassBeanName = beanClassMetadata.getSuperName();
        }

        // if the bean is using this interceptor and that this is not the bean
        // class or parent class, add the link.
        if (classAnnotationMetadata != null && !superClassesOfBean.contains(interceptorClassName)) {
            addInterceptorForClass(classAnnotationMetadata, interceptorClassName);
        }

        M foundMethod = null;
        C interceptorMetadata = classAnnotationMetadata.getLinkedClassMetadata(interceptorClassName);

        boolean loop = true;
        // While method is not found and that class has super class, try to
        // search
        while (foundMethod == null && loop) {
            List<M> listMethods = interceptorMetadata.searchStandardMethodMetadata(methodName);

            // Method present ?
            if (listMethods.size() == 1) {
                // first item
                foundMethod = listMethods.get(0);
                continue;
            } else if (listMethods.size() > 1) {
                throw new IllegalStateException("Method with name '" + methodName + "' was found more than once on the class '"
                        + interceptorMetadata.getClassName() + "'.");
            }

            // else continue
            String superClassName = interceptorMetadata.getSuperName();
            if (!superClassName.equals(JAVA_LANG_OBJECT)) {
                interceptorMetadata = getEjbJarDeployableMetadata().getScannedClassMetadata(superClassName);
            } else {
                // break loop
                loop = false;
            }
        }

        if (foundMethod == null) {
            throw new IllegalStateException("The method named '" + methodName + "' was not found in the class '"
                    + interceptorMetadata.getClassName() + "' and all its super-classes.");
        }

        // call setter to true and only if it is not already present.
        if (POST_CONSTRUCT.equals(type)) {
            if (!foundMethod.isPostConstruct()) {
                foundMethod.setPostConstruct(true);
            }
        } else if (PRE_DESTROY.equals(type)) {
            if (!foundMethod.isPreDestroy()) {
                foundMethod.setPreDestroy(true);
            }
        } else if (PRE_PASSIVATE.equals(type)) {
            if (!foundMethod.isPrePassivate()) {
                foundMethod.setPrePassivate(true);
            }
        } else if (POST_ACTIVATE.equals(type)) {
            if (!foundMethod.isPostActivate()) {
                foundMethod.setPostActivate(true);
            }
        } else if (IAroundInvoke.NAME.equals(type)) {
            if (!foundMethod.isAroundInvoke()) {
                foundMethod.setAroundInvoke(true);
            }
        } else {
            throw new IllegalStateException("No case for type '" + type + "'..");
        }
    }

    /**
     * Set/override the current application-exception.
     * @param assemblyDescriptor the given assembly descriptor to analyze.
     */
    public void mergeApplicationException(final IAssemblyDescriptor assemblyDescriptor) {
        List<IApplicationException> assemblyDescriptorList = assemblyDescriptor.getApplicationExceptionList();
        if (assemblyDescriptorList != null) {
            for (IApplicationException applicationException : assemblyDescriptorList) {
                String className = applicationException.getClassName();
                // Get metadata for the given class
                C exceptionMetadata = getEjbJarDeployableMetadata().getScannedClassMetadata(encode(className));
                if (exceptionMetadata == null) {
                    throw new IllegalStateException("The class named '" + className
                            + "' was not found. Cannot set ApplicationException on it");
                }
                // Set the new value
                exceptionMetadata.setApplicationException(applicationException);
                logger.debug("Setting the application-exception ''{0}'' on the class ''{1}''",
                        applicationException, className);
            }
        }
    }

    /**
     * Set/override the current security.
     * @param assemblyDescriptor the given assembly descriptor to analyze.
     */
    public void mergeAssemblySecurity(final IAssemblyDescriptor assemblyDescriptor) {

        // Search the security role
        List<String> securityRoles = assemblyDescriptor.getSecurityRoleList();
        for (String role : securityRoles) {
            // Gets the metadata for all ejb
            for (C beanMetadata : getEjbJarDeployableMetadata().getEjbJarClassMetadataCollection()) {
                if (beanMetadata.isBean()) {
                    List<String> declaredRoles = beanMetadata.getDeclareRoles();
                    if (declaredRoles == null) {
                        declaredRoles = new ArrayList<String>();
                    }
                    if (!declaredRoles.contains(role)) {
                        declaredRoles.add(role);
                        beanMetadata.setDeclareRoles(declaredRoles);
                    }
                }
            }
        }

        // sets the per-role or unchecked methods
        List<IMethodPermission> methodPermissionList = assemblyDescriptor.getMethodPermissionList();
        for (IMethodPermission methodPermission : methodPermissionList) {

            List<IMethodDD> methodDDs = methodPermission.getMethods();
            for (IMethodDD method : methodDDs) {

                // Get Ejb-Name
                String ejbName = method.getEjbName();

                // Gets the metadata for the ejb
                C beanMetadata = getEjbJarDeployableMetadata().getEjbJarClassMetadataForEjbName(ejbName);
                if (beanMetadata == null) {
                    throw new IllegalStateException("No metadata found for the the ejb '" + ejbName
                            + "'. Check that this ejb is defined. This ejb-name is referenced from "
                            + "a method of an assembly descriptor / method-permission");
                }

                // Get matching methods
                List<M> methods = getMethodsForGivenMethodDD(method, beanMetadata);

                for (M methodAnnotationMetadata : methods) {
                    // Role -base or unchecked ?
                    if (methodPermission.isUnchecked()) {
                        methodAnnotationMetadata.setPermitAll(true);
                    } else if (methodPermission.getRoleNameList().size() > 0) {
                        for (String role : methodPermission.getRoleNameList()) {
                            List<String> existingRolesAllowed = methodAnnotationMetadata.getRolesAllowed();
                            if (existingRolesAllowed == null) {
                                existingRolesAllowed = new ArrayList<String>();
                            }
                            if (!existingRolesAllowed.contains(role)) {
                                existingRolesAllowed.add(role);
                                methodAnnotationMetadata.setRolesAllowed(existingRolesAllowed);
                            }
                        }
                    }
                }
            }
        }

        // Exclude the given methods of the XML DD
        List<IMethodDD> excludeMethodList = assemblyDescriptor.getExcludeListMethods();
        for (IMethodDD method : excludeMethodList) {

            // Get Ejb-Name
            String ejbName = method.getEjbName();

            // Gets the metadata for the ejb
            C beanMetadata = getEjbJarDeployableMetadata().getEjbJarClassMetadataForEjbName(ejbName);
            if (beanMetadata == null) {
                throw new IllegalStateException("No metadata found for the the ejb '" + ejbName
                        + "'. Check that this ejb is defined. This ejb-name is referenced from "
                        + "a method of an assembly descriptor / exclude-list");
            }

            // Get matching methods
            List<M> methods = getMethodsForGivenMethodDD(method, beanMetadata);

            // Set the deny all flag to these methods
            for (M methodAnnotationMetadata : methods) {
                methodAnnotationMetadata.setDenyAll(true);
            }
        }
    }

    /**
     * Set/override the current container-transaction.
     * @param assemblyDescriptor the given assembly descriptor to analyze.
     */
    public void mergeContainerTransaction(final IAssemblyDescriptor assemblyDescriptor) {
        List<IContainerTransaction> containerTransactionList = assemblyDescriptor.getContainerTransactionList();
        if (containerTransactionList == null) {
            return;
        }

        for (IContainerTransaction containerTransaction : containerTransactionList) {

            // No transaction attribute ? then escape
            TransactionAttributeType transactionAttributeType = containerTransaction.getTransactionAttribute();
            if (transactionAttributeType == null) {
                continue;
            }

            for (IMethodDD method : containerTransaction.getMethods()) {
                // Get Ejb-Name
                String ejbName = method.getEjbName();

                // Gets the metadata for the ejb
                C beanMetadata = getEjbJarDeployableMetadata().getEjbJarClassMetadataForEjbName(ejbName);
                if (beanMetadata == null) {
                    throw new IllegalStateException("No metadata found for the the ejb '" + ejbName
                            + "'. Check that this ejb is defined. This ejb-name is referenced from "
                            + "a method of a container transaction");
                }

                // Get Ejb-Name
                String methodName = method.getName();
                TransactionAttributeLevel txlevel = TransactionAttributeLevel.XML_WILDCARD;

                // Style 1 (see �13.3.7.2)
                if (WILDCARD.equals(methodName)) {
                    // Apply tx for all methods if it can be applied
                    for (M methodAnnotationMetadata : beanMetadata.getStandardMethodMetadataCollection()) {
                        // check level ?
                        if (methodAnnotationMetadata.getTransactionAttributeLevel() == TransactionAttributeLevel.ANNOTATION) {
                            methodAnnotationMetadata.setTransactionAttributeType(transactionAttributeType);
                            methodAnnotationMetadata.setTransactionAttributeLevel(txlevel);
                            logger.debug("Setting transaction attribute of method {0} of bean {1} to {2}",
                                    methodAnnotationMetadata, ejbName, transactionAttributeType);
                        }

                    }
                    // next step
                    continue;
                }

                // Style 2 ?
                List<String> params = method.getParams();

                if (params == null || params.size() == 0) {
                    txlevel = TransactionAttributeLevel.XML_METHOD_NAME;
                } else {
                    txlevel = TransactionAttributeLevel.XML_METHOD_PARAMS;
                }

                // Get matching methods
                List<M> methods = getMethodsForGivenMethodDD(method, beanMetadata);
                if (methods.size() == 0) {
                    throw new IllegalStateException("No matching method found with '" + method + "' for bean '" + ejbName
                            + "'.");
                }

                for (M methodAnnotationMetadata : methods) {

                    // check level ?
                    if (methodAnnotationMetadata.getTransactionAttributeLevel() == TransactionAttributeLevel.ANNOTATION
                           || methodAnnotationMetadata.getTransactionAttributeLevel() == TransactionAttributeLevel.XML_WILDCARD) {

                        if ((txlevel == TransactionAttributeLevel.XML_METHOD_NAME && methodAnnotationMetadata
                                .getTransactionAttributeLevel() != TransactionAttributeLevel.XML_METHOD_PARAMS)
                                || (txlevel == TransactionAttributeLevel.XML_METHOD_PARAMS)) {
                            methodAnnotationMetadata.setTransactionAttributeType(transactionAttributeType);
                            methodAnnotationMetadata.setTransactionAttributeLevel(txlevel);
                            logger.debug("Setting transaction attribute of method {0} of bean {1} to {2}",
                                    methodAnnotationMetadata, ejbName, txlevel);
                        }
                    }

                }
            }
        }
    }

    /**
     * Set/override the current interceptor-binding.
     * @param assemblyDescriptor the given assembly descriptor to analyze.
     */
    public void mergeInterceptorBinding(final IAssemblyDescriptor assemblyDescriptor) {
        List<IInterceptorBinding> interceptorBindingList = assemblyDescriptor.getInterceptorBindingList();
        if (interceptorBindingList != null) {
            for (IInterceptorBinding interceptorBinding : interceptorBindingList) {

                // Get Ejb-Name
                String ejbName = interceptorBinding.getEjbName();
                // Default interceptor
                if (WILDCARD.equals(ejbName)) {
                    // Get current list
                    IJInterceptors defaultInterceptors = getEjbJarDeployableMetadata().getDefaultInterceptorsClasses();
                    if (defaultInterceptors == null) {
                        // build new list if empty
                        defaultInterceptors = new JInterceptors();
                        getEjbJarDeployableMetadata().setDefaultInterceptorsClasses(defaultInterceptors);
                    }

                    // Add the interceptors of the interceptor-binding to the
                    // default interceptors
                    for (String interceptor : interceptorBinding.getInterceptorClassList()) {
                        defaultInterceptors.addClass(encode(interceptor));
                    }

                    // stop the loop
                    continue;
                }

                // Gets the metadata for the ejb
                C beanMetadata = getEjbJarDeployableMetadata().getEjbJarClassMetadataForEjbName(ejbName);
                if (beanMetadata == null) {
                    throw new IllegalStateException("No metadata found for the the ejb '" + ejbName
                            + "'. Check that this ejb is defined. This ejb-name is referenced from an interceptor-binding");
                }

                // It is a method ?
                IMethodDD methodDD = interceptorBinding.getMethod();

                // not a method
                if (methodDD == null) {
                    IJInterceptors interceptors = beanMetadata.getAnnotationInterceptors();
                    if (interceptors == null) {
                        // build new list if empty
                        interceptors = new JInterceptors();
                        beanMetadata.setAnnotationsInterceptors(interceptors);
                    }

                    for (String interceptor : interceptorBinding.getInterceptorClassList()) {
                        interceptors.addClass(encode(interceptor));
                    }

                    // Exclude default interceptors for this bean ?
                    if (interceptorBinding.isExcludeDefaultInterceptorsCalled()) {
                        beanMetadata.setExcludeDefaultInterceptors(interceptorBinding.isExcludeDefaultInterceptors());
                    }

                    // For ordered item: First, add interceptors (if they are
                    // not already present)
                    List<String> orderList = interceptorBinding.getOrderInterceptorClassList();
                    if (orderList != null) {

                        // Check that this is the first time that we try to
                        // order the list.
                        if (beanMetadata.isOrderedInterceptors()) {
                            throw new IllegalStateException("The interceptor-order element has already been used on the bean '"
                                    + ejbName + "'. It can only be used once.");
                        }

                        // Flag class as a class having ordered interceptors
                        beanMetadata.setOrderedInterceptors(true);

                        // Build a list with all interceptors. (it means that we
                        // add default interceptors if any) if they are not
                        // excluded
                        if (!beanMetadata.isExcludedDefaultInterceptors()) {
                            if (beanMetadata.getEjbJarDeployableMetadata().getDefaultInterceptorsClasses() != null) {
                                for (String cls : beanMetadata.getEjbJarDeployableMetadata().getDefaultInterceptorsClasses()
                                        .getClasses()) {
                                    interceptors.addClass(cls);
                                }
                            }
                        }

                        // Then add new items found in the DD
                        for (String interceptor : interceptorBinding.getOrderInterceptorClassList()) {
                            String encodedName = encode(interceptor);
                            if (!interceptors.contains(encodedName)) {
                                interceptors.addClass(encodedName);
                            }
                        }

                        // Order.
                        // First, check that the list has the correct size as it
                        // should be a total ordering.
                        if (interceptors.size() != orderList.size()) {
                            String err = "The list used for ordering interceptors is not a total list as the size mismatch.";
                            err += "The current list is '" + interceptors + "' while the ordering list is '" + orderList + "'.";
                            throw new IllegalStateException(err);
                        }

                        // Size is ok, check that all elements are present
                        for (String clazz : orderList) {
                            if (!interceptors.contains(encode(clazz))) {
                                throw new IllegalStateException("The element '" + clazz
                                        + "' of the ordered list is not present in the interceptors '" + interceptors + "'.");
                            }
                        }

                        // Ok, all the elements are present, set the new
                        // ordering.
                        // Which is based on the given ordered list.
                        JInterceptors orderedInterceptors = new JInterceptors();
                        for (String clazz : orderList) {
                            orderedInterceptors.addClass(encode(clazz));
                        }
                        // Set the new list
                        beanMetadata.setAnnotationsInterceptors(orderedInterceptors);

                    }
                } else {
                    // method case
                    List<M> methods = getMethodsForGivenMethodDD(methodDD, beanMetadata);

                    // If no elements, print warning
                    if (methods.size() == 0) {
                        logger.warn("Method with name '" + methodDD.getName()
                                + "' is specified in interceptorBinding element for class '" + beanMetadata.getClassName()
                                + "' but no methods are matching");
                    }

                    // now, apply interceptor on each method
                    for (M method : methods) {
                        IJInterceptors interceptors = method.getAnnotationInterceptors();
                        if (interceptors == null) {
                            // build new list if empty
                            interceptors = new JInterceptors();
                            method.setAnnotationsInterceptors(interceptors);
                        }

                        for (String interceptor : interceptorBinding.getInterceptorClassList()) {
                            interceptors.addClass(encode(interceptor));
                        }
                        
                        // Then add ordered interceptors if any
                        List<String> orderList = interceptorBinding.getOrderInterceptorClassList();
                        if (orderList != null) {
                            for (String interceptor : interceptorBinding.getOrderInterceptorClassList()) {
                                String encodedName = encode(interceptor);
                                if (!interceptors.contains(encodedName)) {
                                    interceptors.addClass(encodedName);
                                }
                            }
                        }

                        // Exclude class interceptors for this method ?
                        if (interceptorBinding.isExcludeClassInterceptorsCalled()) {
                            method.setExcludeClassInterceptors(interceptorBinding.isExcludeClassInterceptors());
                        }
                        // Exclude default interceptors for this method ?
                        if (interceptorBinding.isExcludeDefaultInterceptorsCalled()) {
                            method.setExcludeDefaultInterceptors(interceptorBinding.isExcludeDefaultInterceptors());
                        }

                    }
                }
            }
        }
    }

    /**
     * Gets the methods metadata that are matching a given method DD.
     * @param methodDD the methodDD object
     * @param classAnnotationMetadata the class where to search methods.
     * @return the list of methods metadata
     */
    protected List<M> getMethodsForGivenMethodDD(final IMethodDD methodDD,
            final C classAnnotationMetadata) {
        // get the method.
        String methodName = methodDD.getName();
        List<String> methodParams = methodDD.getParams();

        // get all methods in the class and its super class
        Map<JMethod, M> beanMethods = new HashMap<JMethod, M>();

        // get super class methods
        String superClassName = classAnnotationMetadata.getClassName();
        while (!"java/lang/Object".equals(superClassName)) {
            C superClassAnnotationMetadata = classAnnotationMetadata.getLinkedClassMetadata(superClassName);
            List<M> findingMethods = superClassAnnotationMetadata.searchStandardMethodMetadata(methodName);
            // add method if not already present
            for (M method : findingMethods) {
                if (!beanMethods.containsKey(method.getJMethod())) {
                    Type[] argumentTypes = Type.getArgumentTypes(method.getJMethod().getDescriptor());
                    // Check that argument of the method are the same than the method found
                    if (methodParams != null) {
                        // Specif params specified in the DD, check the parameters
                        if (argumentTypes != null && argumentTypes.length == methodParams.size()) {
                            beanMethods.put(method.getJMethod(), method);
                        }                        
                    } else {
                        // No parameters specified in the DD, all methods are matching
                        beanMethods.put(method.getJMethod(), method);
                    }
                }
            }
            superClassName = superClassAnnotationMetadata.getSuperName();
        }

        // Found methods are values of the map
        List<M> methods = new ArrayList<M>(beanMethods.values());

        // params ?
        List<String> params = methodDD.getParams();
        // Keep only method with the correct params
        if (params != null && params.size() > 0) {
            List<M> newMethods = new ArrayList<M>();
            for (M method : methods) {
                // get params for the current method
                Type[] types = Type.getArgumentTypes(method.getJMethod().getDescriptor());
                // Same size ?
                if (params.size() != types.length) {
                    // Not same size, go to the next one
                    continue;
                }

                // Compare each element
                int i = 0;
                boolean matching = false;
                for (String param : params) {
                    if (param.equals(types[i].getClassName())) {
                        matching = true;
                    }
                    i++;
                }
                // match ?
                if (matching) {
                    newMethods.add(method);
                }
            }
            // Get the new list (only matching elements)
            methods = newMethods;
        }
        return methods;
    }

    /**
     * Merge the @HandlerChain annotations in their colocated @WebServiceRef
     */
    protected void applyHandlerChain(final C classAnnotationMetadata) {

        // Browse HandlerChains and @WebServiceRef supported by classes
        IJwsHandlerChain handlerChain = classAnnotationMetadata.getJwsHandlerChain();
        List<IJaxwsWebServiceRef> references = classAnnotationMetadata.getJaxwsWebServiceRefs();

        if ((handlerChain != null) && (references != null) && (!references.isEmpty())) {
            // We have a HandlerChain and some @WebServiceRef annotations
            for (IJaxwsWebServiceRef reference : references) {
                reference.setHandlerChainFile(handlerChain.getFile());

                // Store declaring class
                String classname = decode(classAnnotationMetadata.getJClass().getName());
                reference.setDeclaringClass(classname);
            }
        }

        // Browse HandlerChains and @WebServiceRef supported by methods
        Collection<? extends ICommonMethodMetadata> allMethodMetadata = classAnnotationMetadata.getStandardMethodMetadataCollection();
        for (ICommonMethodMetadata metadata : allMethodMetadata) {

            IJwsHandlerChain jwsHandlerChain = metadata.getJwsHandlerChain();
            IJaxwsWebServiceRef reference = metadata.getJaxwsWebServiceRef();

            if ((jwsHandlerChain != null) && (reference != null)) {
                // We have a HandlerChain and some @WebServiceRef annotations
                reference.setHandlerChainFile(jwsHandlerChain.getFile());

                // Store declaring class
                String classname = decode(metadata.getClassMetadata().getJClass().getName());
                reference.setDeclaringClass(classname);
            }
        }


        // Browse HandlerChains and @WebServiceRef supported by fields
        Collection<? extends ICommonFieldMetadata> allFieldMetadatas = classAnnotationMetadata.getStandardFieldMetadataCollection();
        for (ICommonFieldMetadata metadata : allFieldMetadatas) {

            IJwsHandlerChain jwsHandlerChain = metadata.getJwsHandlerChain();
            IJaxwsWebServiceRef reference = metadata.getJaxwsWebServiceRef();

            if ((jwsHandlerChain != null) && (reference != null)) {
                // We have a HandlerChain and some @WebServiceRef annotations
                reference.setHandlerChainFile(jwsHandlerChain.getFile());

                // Store declaring class
                String classname = decode(metadata.getClassMetadata().getJClass().getName());
                reference.setDeclaringClass(classname);
            }
        }
    }
}
